package cz.muni.fi.gravemanager.backend;


import cz.muni.fi.pv168.gravemanager.backend.Grave;
import cz.muni.fi.pv168.gravemanager.backend.GraveManager;
import cz.muni.fi.pv168.gravemanager.backend.GraveManagerImpl;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import static org.junit.Assert.*;


public class GraveManagerImplTest {

    private GraveManager manager;

    @Before
    public void setUp() throws SQLException {
        manager = new GraveManagerImpl();
    }

    @Test
    public void createGrave() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        assertNotNull(graveId);

        Grave result = manager.getGrave(graveId);
        assertEquals(grave, result);
        assertNotSame(grave, result);
        assertDeepEquals(grave, result);
    }

    @Test
    public void getGrave() {
        assertNull(manager.getGrave(1L));

        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);
        Long graveId = grave.getId();

        Grave result = manager.getGrave(graveId);
        assertEquals(grave, result);
        assertDeepEquals(grave, result);
    }

    @Test
    public void getAllGraves() {
        assertTrue(manager.findAllGraves().isEmpty());

        Grave g1 = newGrave(23, 44, 5, "Grave 1");
        Grave g2 = newGrave(12, 4, 1,"Grave 2");

        manager.createGrave(g1);
        manager.createGrave(g2);

        List<Grave> expected = Arrays.asList(g1,g2);
        List<Grave> actual = manager.findAllGraves();

        Collections.sort(actual, idComparator);
        Collections.sort(expected, idComparator);

        assertEquals(expected, actual);
        assertDeepEquals(expected, actual);
    }



    //the following three variants of the method addNullGrave show different ways to handle required exceptions
    //(in the rest of the testing methods, we use them interchangeably)
    @Test
    public void addNullGrave_1() {
        try {
            manager.createGrave(null);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }
    }

    @Test(expected = IllegalArgumentException.class)
    public void addNullGrave_2() {
        manager.createGrave(null);
    }

    //the Rule should probably be defined at the top in the 'attributes section'; we put it here just to show where it is used
    @Rule
    public final ExpectedException exception = ExpectedException.none();

    @Test
    public void addNullGrave_3() {
        //If this was a more complex test, here could be some code that should not (but potentially _could_) throw IllegalArgExc.
        // If it does, the test fails. If we used the annotation (@Test(expected....), it would apply to the whole method, so
        // the test would pass regardless of which statement threw the exception.
        exception.expect(IllegalArgumentException.class); //from now on, if IllegalArgExc is thrown, it is ok
        manager.createGrave(null);
    }



    @Test
    public void addGraveWithAssignedId() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        grave.setId(1L);
        try {
            manager.createGrave(grave);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }
    }

    @Test(expected = IllegalArgumentException.class)
    public void addGraveWithNegativeColumn() {
        Grave grave = newGrave(-1, 13, 6, "Nice grave");
        manager.createGrave(grave);
    }

    @Test(expected = IllegalArgumentException.class)
    public void addGraveWithNegativeRow() {
        Grave grave = newGrave(1, -1, 6, "Nice grave");
        manager.createGrave(grave);
    }

    @Test(expected = IllegalArgumentException.class)
    public void addGraveWithNegativeCapacity() {
        Grave grave = newGrave(1, 1, -1, "Nice grave");
        manager.createGrave(grave);
    }

    @Test(expected = IllegalArgumentException.class)
    public void addGraveWithZeroCapacity() {
        Grave grave = newGrave(1, 1, 0, "Nice grave");
        manager.createGrave(grave);
    }

    @Test
    public void addGraveWithColumnZero() {
        //should be ok
        Grave grave = newGrave(0, 13, 6, "Nice grave");
        manager.createGrave(grave);
        Grave result = manager.getGrave(grave.getId());
        assertNotNull(result);
    }

    @Test
    public void addGraveWithRowZero() {
        //should be ok
        Grave grave = newGrave(12, 0, 6, "Nice grave");
        manager.createGrave(grave);
        Grave result = manager.getGrave(grave.getId());
        assertNotNull(result);
    }

    @Test
    public void addGraveOK() {
        Grave grave = newGrave(12, 11, 6, null);
        manager.createGrave(grave);
        Grave result = manager.getGrave(grave.getId());
        assertNotNull(result);
        assertNull(result.getNote());
    }

    @Test
    public void updateGrave() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        Grave g2 = newGrave(18, 19, 100, "Another record");
        manager.createGrave(grave);
        manager.createGrave(g2);

        Long graveId = grave.getId();
        Grave graveResult = manager.getGrave(graveId);

        graveResult.setColumn(0);
        manager.updateGrave(graveResult);
        graveResult = manager.getGrave(graveId);
        assertEquals(0, graveResult.getColumn());
        assertEquals(13, graveResult.getRow());
        assertEquals(6, graveResult.getCapacity());
        assertEquals("Nice grave", graveResult.getNote());

        graveResult.setRow(0);
        manager.updateGrave(graveResult);
        graveResult = manager.getGrave(graveId);
        assertEquals(0, graveResult.getColumn());
        assertEquals(0, graveResult.getRow());
        assertEquals(6, graveResult.getCapacity());
        assertEquals("Nice grave", graveResult.getNote());

        graveResult.setCapacity(1);
        manager.updateGrave(graveResult);
        graveResult = manager.getGrave(graveId);
        assertEquals(0, graveResult.getColumn());
        assertEquals(0, graveResult.getRow());
        assertEquals(1, graveResult.getCapacity());
        assertEquals("Nice grave", graveResult.getNote());

        graveResult.setNote("Another grave");
        manager.updateGrave(graveResult);
        graveResult = manager.getGrave(graveId);
        assertEquals(0, graveResult.getColumn());
        assertEquals(0, graveResult.getRow());
        assertEquals(1, graveResult.getCapacity());
        assertEquals("Another grave", graveResult.getNote());

        graveResult.setNote(null);
        manager.updateGrave(graveResult);
        graveResult = manager.getGrave(graveId);
        assertEquals(0, graveResult.getColumn());
        assertEquals(0, graveResult.getRow());
        assertEquals(1, graveResult.getCapacity());
        assertNull(graveResult.getNote());

        //make sure the updates didn't affect other records
        assertDeepEquals(g2, manager.getGrave(g2.getId()));
    }

    @Test(expected = IllegalArgumentException.class)
    public void updateNullGrave() {
        manager.updateGrave(null);
    }

    @Test
    public void updateGraveSetIdNull() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        grave = manager.getGrave(graveId);

        grave.setId(null);

        exception.expect(IllegalArgumentException.class);
        manager.updateGrave(grave);
    }

    @Test
    public void updateGraveSetIdNegative() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        grave = manager.getGrave(graveId);

        grave.setId(-1L);

        exception.expect(IllegalArgumentException.class);
        manager.updateGrave(grave);
    }

    @Test
    public void updateGraveSetColumnNegative() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        grave = manager.getGrave(graveId);

        grave.setColumn(-1);

        exception.expect(IllegalArgumentException.class);
        manager.updateGrave(grave);
    }

    @Test
    public void updateGraveSetRowNegative() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        grave = manager.getGrave(graveId);

        grave.setRow(-1);

        exception.expect(IllegalArgumentException.class);
        manager.updateGrave(grave);
    }

    @Test
    public void updateGraveSetCapacityZero() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        grave = manager.getGrave(graveId);

        grave.setCapacity(0);

        exception.expect(IllegalArgumentException.class);
        manager.updateGrave(grave);
    }


    @Test
    public void updateGraveSetCapacityNegative() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");
        manager.createGrave(grave);

        Long graveId = grave.getId();
        grave = manager.getGrave(graveId);

        grave.setCapacity(-1);

        exception.expect(IllegalArgumentException.class);
        manager.updateGrave(grave);
    }

    @Test
    public void deleteGrave() {
        Grave g1 = newGrave(12, 13, 6, "Nice grave");
        Grave g2 = newGrave(18, 19, 100, "Another record");
        manager.createGrave(g1);
        manager.createGrave(g2);

        assertNotNull(manager.getGrave(g1.getId()));
        assertNotNull(manager.getGrave(g2.getId()));

        manager.deleteGrave(g1);

        assertNull(manager.getGrave(g1.getId()));
        assertNotNull(manager.getGrave(g2.getId()));
    }

    @Test(expected = IllegalArgumentException.class)
    public void deleteNullGrave() {
        manager.deleteGrave(null);
    }

    @Test
    public void deleteGraveWithNullId() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");

        grave.setId(null);

        try {
            manager.deleteGrave(grave);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }
    }

    @Test
    public void deleteGraveWithNonexistentId() {
        Grave grave = newGrave(12, 13, 6, "Nice grave");

        grave.setId(100L);

        try {
            manager.deleteGrave(grave);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }
    }


    private static Grave newGrave(int column, int row, int capacity, String note) {
        Grave grave = new Grave();
        grave.setColumn(column);
        grave.setRow(row);
        grave.setCapacity(capacity);
        grave.setNote(note);
        return grave;
    }

    private void assertDeepEquals(List<Grave> expectedList, List<Grave> actualList) {
        for (int i = 0; i < expectedList.size(); i++) {
            Grave expected = expectedList.get(i);
            Grave actual = actualList.get(i);
            assertDeepEquals(expected, actual);
        }
    }

    private void assertDeepEquals(Grave expected, Grave actual) {
        assertEquals(expected.getId(), actual.getId());
        assertEquals(expected.getColumn(), actual.getColumn());
        assertEquals(expected.getRow(), actual.getRow());
        assertEquals(expected.getCapacity(), actual.getCapacity());
        assertEquals(expected.getNote(), actual.getNote());
    }

    private static Comparator<Grave> idComparator = new Comparator<Grave>() {

        @Override
        public int compare(Grave o1, Grave o2) {
            return o1.getId().compareTo(o2.getId());
        }

    };
    //Java 8 lambdas:
    // private static Comparator<Grave> idComparator = (o1, o2) -> o1.getId().compareTo(o2.getId());

}
